import asyncio
import importlib
import inspect
import json
import subprocess
import time
from asyncio import Task, Event
from importlib import machinery, util

from aiohttp import ClientSession, ClientTimeout

import tordl.config as cfg


def run_torrent_client(magnet_url):
    cmd = (cfg.TORRENT_CLIENT_CMD % magnet_url).split(' ')
    subprocess.Popen(
        cmd,
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL
    )


def open_torrent_link(link):
    cmd = (cfg.BROWSER_CMD % link).split(' ')
    subprocess.Popen(
        cmd,
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL
    )


def direct_download(st, loop=None):
    if not st:
        print('No search term defined, cannot use --download option.')
        exit(1)

    if not loop:
        loop = asyncio.get_event_loop()

    dl = DlFacade()
    print('Searching %s for "%s"...' % (','.join(cfg.SEARCH_ENGINES), st))
    results = loop.run_until_complete(dl.search(st))
    if results:
        results = sorted(results, key=lambda r: r.seeders, reverse=True)
        result = results[0]
        print('Found "%s", seeds: %s, size: %s,  ' % (
            result.name, result.seeders, result.size
        ))
        if result.magnet_url:
            magnet_url = result.magnet_url
        else:
            print('Fetching magnet link...')
            magnet_url = loop.run_until_complete(dl.get_magnet_url(result))

        print('Running torrent client...')
        run_torrent_client(magnet_url)
    else:
        print('No results found.')


def test_search_engines(loop=None):
    if not loop:
        loop = asyncio.get_event_loop()

    st = 'Star Wars'
    dl = DlFacade()
    es = dl.engines.values()
    test_results = []
    for e in es:
        cls = e.__class__
        print(
            'Testing %s [%s]...' % (
                '%s.%s' % (cls.__module__, cls.__qualname__), e.NAME
            )
        )
        error = False
        t = time.time()
        results = loop.run_until_complete(e.search(st))
        t = time.time() - t
        if len(results) > 0:
            print('  - Search results (%d) fetched.' % len(results))
            r = results[0]
            if not r.name:
                error = True
                print('  - ERROR: Name not found !')
            if not r.link and not r.magnet_url:
                error = True
                print('  - ERROR: Link not found !')
            if not r.seeders:
                error = True
                print('  - ERROR: Seeders not found !')
            if not r.leechers:
                error = True
                print('  - ERROR: Leechers not found !')
            if not r.size:
                error = True
                print('  - ERROR: Size not found !')
            if not r.link:
                error = True
                print('  - ERROR: Link not found !')
            if not r.magnet_url:
                magnet_url = loop.run_until_complete(e.get_magnet_url(r))
                if magnet_url:
                    print('  - Magnet URL fetched.')
                else:
                    error = True
                    print('  - ERROR: MagnetURL not found !')
        else:
            error = True
            print('  - ERROR, no results found !')
        if not error:
            test_results.append('[OK] %s [search_time=%.3fs]' % (e.NAME, t))
        else:
            test_results.append('[ERR] %s [search_time=%.3fs]' % (e.NAME, t))

    print('-' * 20)
    for m in test_results:
        print(m)
    print('-' * 20)


def run_api(st, pretty_json=True, loop=None):
    if not st:
        print('No search term defined, cannot use --api option.')
        exit(1)

    if not loop:
        loop = asyncio.get_event_loop()

    api = Api()
    sr = loop.run_until_complete(api.fetch(st))
    print(api.mk_json_output(sr, pretty_json))


class SearchResult(object):
    def __init__(
            self, origin, name, link, seeders, leechers, size, magnet_url=None
    ):
        self.origin = origin
        self.name = name.encode('ascii', 'ignore').decode()
        self.link = link
        self.seeders = seeders
        self.leechers = leechers
        self.size = size.replace(' ', '').encode('ascii', 'ignore').decode()
        self.magnet_url = magnet_url

        sb = self.size.lower()
        if 'kb' in sb:
            self.size_b = float(sb.replace('kb', '')) * 1024
        elif 'mb' in sb:
            self.size_b = float(sb.replace('mb', '')) * (1024 ** 2)
        elif 'gb' in sb:
            self.size_b = float(sb.replace('gb', '')) * (1024 ** 3)
        elif 'tb' in sb:
            self.size_b = float(sb.replace('tb', '')) * (1024 ** 4)
        elif 'b' in sb:
            self.size_b = float(sb.replace('b', ''))
        else:
            try:
                self.size_b = float(sb)
            except Exception:
                pass
            else:
                self.size_b = 0.0


class BaseDl(object):
    NAME = ''
    BASE_URL = None
    SEARCH_URL = None
    INDEXED = True

    def __init__(self):
        self._current_index = 1
        self._current_search = None
        self._headers = self._create_headers()

    async def search(self, expression, new_search=False):
        if expression is None and not new_search:
            if not self.INDEXED:
                return []

            self._current_index += 1
        else:
            self._current_index = 1

        if expression:
            self._current_search = expression

        self._set_referer('%s/' % self.BASE_URL)
        response = await self._get_url(
            self._mk_search_url(self._current_search)
        )

        return self._process_search(response) if response else None

    async def get_magnet_url(self, search_result):
        self._set_referer(self._mk_search_url(self._current_search))
        response = await self._get_url(self._mk_magnet_url(search_result.link))
        return self._process_magnet_link(response) if response else None

    async def _get_url(self, url):
        try:
            async with ClientSession(
                    headers=self._headers,
                    timeout=ClientTimeout(cfg.REQUEST_TIMEOUT)
            ) as sess:
                async with sess.get(url) as response:
                    return await response.read()
        except asyncio.exceptions.TimeoutError:
            return None

    def _mk_search_url(self, expression):
        raise NotImplementedError()

    def _mk_magnet_url(self, link):
        raise NotImplementedError()

    def _process_search(self, response):
        raise NotImplementedError()

    def _process_magnet_link(self, response):
        raise NotImplementedError()

    def _create_headers(self):
        return {
            'Accept': 'text/html,application/xhtml+xml,'
                      'application/xml;q=0.9,image/webp,'
                      '*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'en-US,en;q=0.5',
            'Connection': 'keep-alive',
            'Content-Type': 'application/x-www-form-urlencoded',
            'Host': self.BASE_URL.split('://')[1],
            'Origin': self.BASE_URL,
            'Set-GPC': '1',
            'Upgrade-Insecure-Requests': '1',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64;'
                          ' x64) AppleWebKit/537.36 (KHTML, '
                          'like Gecko) Chrome/90.0.4430.72 '
                          'Safari/537.36'
        }

    def _set_referer(self, url):
        self._headers['Referer'] = url


class DlFacade(object):

    def __init__(
            self,
            dl_classes=None,
    ):
        if dl_classes:
            self._engines = {c: c() for c in dl_classes}
            self._all_engines = self._load_engines()[1]
        else:
            self._engines, self._all_engines = self._load_engines()

    @property
    def engines(self):
        return self._engines

    @property
    def all_engines(self):
        return self._all_engines

    async def search(self, expression, new_search=False):
        tasks = []
        for dl in self._engines.values():
            tasks.append(dl.search(expression, new_search))

        results, _ = await asyncio.wait(tasks)
        result = []
        for r in results:
            r = r.result()
            if r:
                result.extend(r)

        return result

    async def fetch_pages(self, search_term):
        coros = []
        for i in range(cfg.PAGE_NUM_DOWNLOAD):
            coros.append(
                self.search(search_term if i == 0 else None)
            )
        done, _ = await asyncio.wait(coros)
        items = []
        for t in done:
            result = t.result()
            if result:
                items.extend(result)

        return items

    async def get_magnet_url(self, search_result):
        e = self._engines[search_result.origin]
        return await e.get_magnet_url(search_result)

    def _load_engines(self):
        loader = importlib.machinery.SourceFileLoader(
            'engines_mod', cfg.CFG_ENGINES_FILE
        )
        spec = importlib.util.spec_from_loader('engines_mod', loader)
        engines_mod = importlib.util.module_from_spec(spec)
        loader.exec_module(engines_mod)

        all_engines = []
        for name, obj in inspect.getmembers(engines_mod):
            if inspect.isclass(obj):
                mro = obj.mro()
                if len(mro) > 2 and BaseDl in mro:
                    all_engines.append(obj)

        engines = {}
        for c in all_engines:
            if c.NAME in cfg.SEARCH_ENGINES:
                engines[c] = c()

        if not engines:
            raise RuntimeError("No search engines selected.")

        return engines, all_engines


class Api(object):
    class GetMagnetUrlTask(Task):
        def __init__(self, coro, search_result, max_):
            super().__init__(coro)
            self.max_ = max_
            self.search_result = search_result

    def __init__(self, dl_classes=None, concurrent=20):
        self._dl = DlFacade(dl_classes)
        self._concurrent = concurrent

        self._counter = 0
        self._batch_done = Event()

    async def fetch(self, search_term):
        search_results = await self._dl.fetch_pages(search_term)

        await self._fetch_magnet_links(search_results)
        return search_results

    def mk_json_output(self, search_results, pretty=False):
        result = []
        j = {'result': result}
        for sr in search_results:
            o = {
                'name': sr.name,
                'link': '%s%s' % (sr.origin.BASE_URL, sr.link),
                'magnet_url': sr.magnet_url,
                'origin': sr.origin.NAME,
                'seeds': sr.seeders,
                'leeches': sr.leechers,
                'size': sr.size
            }
            result.append(o)

        return json.dumps(j, indent=4 if pretty else None)

    async def _fetch_magnet_links(self, search_results):
        no_ml = []
        for sr in search_results:
            if not sr.magnet_url:
                no_ml.append(sr)

        ln = len(no_ml)
        for i in range(0, ln, self._concurrent):
            max_ = min(i + self._concurrent, ln)
            self._counter = 0
            self._batch_done.clear()
            for j in range(i, max_):
                sr = no_ml[j]
                task = self.GetMagnetUrlTask(
                    self._dl.get_magnet_url(sr), sr, max_ - i
                )
                task.add_done_callback(self._on_fetch_magnet_done)

            await self._batch_done.wait()

    def _on_fetch_magnet_done(self, task):
        self._counter += 1
        task.search_result.magnet_url = task.result()
        if self._counter >= task.max_:
            self._batch_done.set()



import json
import os
import sys

from tordl import engines

CFG_DIR = os.path.join(os.path.expanduser('~'), '.torrentdl')
CFG_FILE = os.path.join(CFG_DIR, 'config.json')
CFG_ENGINES_FILE = os.path.join(CFG_DIR, 'engines.py')
CFG_HISTORY_FILE = os.path.join(CFG_DIR, 'search_history.txt')

SEARCH_ENGINES = ['Zooqle', 'TPB', 'Lime', '1337x', 'Glo', 'KAT', 'Solid']
TORRENT_CLIENT_CMD = 'qbittorrent %s'
BROWSER_CMD = 'firefox %s'
HISTORY_MAX_LENGTH = 100
PAGE_NUM_DOWNLOAD = 1
REQUEST_TIMEOUT = 5


def mk_cfg():
    mod = sys.modules[__name__]
    attrs = dir(mod)
    omit = (
        'os',
        'json',
        'sys',
        'mk_cfg',
        'init_cfg',
        'override_cfg',
        'write_cfg',
        'engines'
    )
    config = {}
    for a in attrs:
        if not a.startswith('CFG') and not a.startswith('__') and a not in omit:
            config[a.lower()] = getattr(mod, a)

    return config


def write_cfg():
    with open(CFG_FILE, 'w') as f:
        f.write(json.dumps(mk_cfg(), indent=4))


def init_cfg():
    if not os.path.exists(CFG_DIR):
        os.makedirs(CFG_DIR)

    if not os.path.exists(CFG_FILE):
        write_cfg()

    if not os.path.exists(CFG_ENGINES_FILE):
        with open(engines.__file__) as f:
            engines_module = f.read()

        with open(CFG_ENGINES_FILE, 'w') as f:
            f.write(engines_module)

    with open(CFG_FILE) as f:
        config = json.load(f)

    mod = sys.modules[__name__]
    for k, v in config.items():
        setattr(mod, k.upper(), v)


def override_cfg(args):
    mod = sys.modules[__name__]
    mod.SEARCH_ENGINES = args.cfg_search_engines.replace(' ', '').split(',')

    omit = ('cfg_search_engines', 'tordl')
    prefix = 'cfg_'
    for k in args.__dict__:
        if k.startswith(prefix) and k not in omit:
            setattr(mod, k.upper()[len(prefix):], getattr(args, k))



#!/usr/bin/env python3

import argparse
import curses
import os
import shutil
import sys
from functools import partial

from tordl import config as cfg, core
from tordl.app import App


def parse_args():
    class ArgParseFormatter(argparse.ArgumentDefaultsHelpFormatter):
        def _get_default_metavar_for_optional(self, action):
            return ''

    ap = argparse.ArgumentParser(
        description='CLI Torrent Downloader provides convenient and quick way '
                    'to search torrent magnet links (and to run associated '
                    'torrent client) via major torrent sites (ThePirateBay, '
                    'LimeTorrents, Zooqle, 1337x, GloTorrents, KickAssTorrents '
                    'by default) through command line.',
        formatter_class=ArgParseFormatter
    )
    ap.add_argument(
        'search',
        nargs='*',
        default='',
        help='Search term. All positional arguments are concatenated to one '
             'string, no need for " "'
    )
    ap.add_argument(
        '-r',
        '--revert-to-default',
        action='store_true',
        default=False,
        help='Purge all custom configuration and revert to default.'
    )
    ap.add_argument(
        '-e',
        '--search-engines',
        dest='cfg_search_engines',
        default=','.join(cfg.SEARCH_ENGINES),
        help='Search engines to be used for torrent search. Overrides the value'
             ' loaded from config in %s' % cfg.CFG_FILE
    )
    ap.add_argument(
        '-c',
        '--torrent-client-cmd',
        dest='cfg_torrent_client_cmd',
        default=cfg.TORRENT_CLIENT_CMD,
        help='Command to execute torrent client with magnet link as a parameter'
    )
    ap.add_argument(
        '-b',
        '--browser-cmd',
        dest='cfg_browser_cmd',
        default=cfg.BROWSER_CMD,
        help='Command to open torrent link in a browser.'
    )
    ap.add_argument(
        '-d',
        '--download',
        action='store_true',
        default=False,
        help='Directly download first search result, don\'t run UI.'
    )
    ap.add_argument(
        '-t',
        '--test-search-engines',
        action='store_true',
        default=False,
        help='Test all active search engines for errors.'
    )
    ap.add_argument(
        '-n',
        '--page-num-download',
        dest='cfg_page_num_download',
        default=cfg.PAGE_NUM_DOWNLOAD,
        type=int,
        help='Fetch N pages of search results from search engines.'
    )
    ap.add_argument(
        '-a',
        '--api',
        default=False,
        action='store_true',
        help='Run in API mode: fetch result and print json to the stdout.'
    )
    ap.add_argument(
        '-p',
        '--pretty-json',
        default=False,
        action='store_true',
        help='Print JSON in pretty format if using --api mode.'
    )
    ap.add_argument(
        '-o',
        '--timeout',
        dest='cfg_request_timeout',
        default=cfg.REQUEST_TIMEOUT,
        type=float,
        help='Search / fetch magnet URL request timeout.'
    )
    parsed = ap.parse_args(sys.argv[1:])
    return parsed


def run_curses_ui(st):
    os.environ.setdefault('ESCDELAY', '0')
    curses.wrapper(partial(App, search=st))


if __name__ == "__main__":
    if '-r' in sys.argv or '--revert-to-default' in sys.argv:
        if os.path.exists(cfg.CFG_DIR):
            shutil.rmtree(cfg.CFG_DIR)
        cfg.init_cfg()
        print('Reverted to default config.')
        exit(0)

    cfg.init_cfg()
    parsed_args = parse_args()
    cfg.override_cfg(parsed_args)

    search_term = ' '.join(parsed_args.search)

    if parsed_args.download:
        core.direct_download(search_term)
    elif parsed_args.test_search_engines:
        core.test_search_engines()
    elif parsed_args.api:
        core.run_api(search_term, parsed_args.pretty_json)
    else:
        run_curses_ui(search_term)

    exit(0)



from distutils.core import setup

setup(
    name='tordl',
    packages=['tordl'],
    version='1.04',
    license='MIT',
    description='CLI Torrent Search and Download',
    author='x0r0x',
    author_email='jakub.schimer@protonmail.com',
    url='https://github.com/X0R0X/cli-torrent-dl/',
    download_url='https://github.com/X0R0X/cli-torrent-dl/archive/refs/heads/master.zip',
    keywords=['torrent', 'search', 'download', 'cli', 'curses'],
    install_requires=[
        'aiohttp==3.7.4.post0',
        'async-timeout==3.0.1',
        'attrs==20.3.0',
        'beautifulsoup4==4.9.3',
        'bs4==0.0.1',
        'chardet==4.0.0',
        'idna==3.1',
        'multidict==5.1.0',
        'pkg-resources==0.0.0',
        'soupsieve==2.2.1',
        'typing-extensions==3.7.4.3',
        'yarl==1.6.3',
        'aiohttp==3.7.4.post0',
        'beautifulsoup4',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Developers',
        'Topic :: Communications :: File Sharing',
        'Topic :: Home Automation',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Terminals',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
)



import asyncio
import curses
import os
import string
from collections import deque
from curses import ascii

from tordl import config as cfg
from tordl import core
from tordl.core import SearchResult, DlFacade


class TopBar(object):
    NO_CAPTION = 'No.'
    TITLE_CAPTION = 'Title'
    SEED_CAPTION = 'Seed'
    LEECH_CAPTION = 'Leech'
    SIZE_CAPTION = 'Size'
    SOURCE_CAPTION = 'Source'

    def __init__(self, screen):
        self._screen = screen

        self._window_top = screen.subwin(1, 0, 0, 1)
        self._window_top.bkgd(' ', curses.color_pair(1) | curses.A_BOLD)

    def resize(self):
        self._window_top = self._screen.subwin(1, 0, 0, 1)
        self._window_top.bkgd(' ', curses.color_pair(1) | curses.A_BOLD)
        self._window_top.clear()

    def draw(
            self, w, no_len, source_max, seeders_max, leechers_max, size_max
    ):
        self._window_top.addstr(0, 0, ' ' * (w - 2))
        self._window_top.addstr(0, 0, self.NO_CAPTION)
        x = no_len + 2
        self._window_top.addstr(0, x, self.TITLE_CAPTION)
        x = (w - (seeders_max + leechers_max + size_max + source_max)) - len(
            self.NO_CAPTION
        ) - 1
        self._window_top.addstr(0, x, self.SOURCE_CAPTION)
        x += source_max + 1
        self._window_top.addstr(0, x, self.SEED_CAPTION)
        x += seeders_max + 1
        self._window_top.addstr(0, x, self.LEECH_CAPTION)
        x += leechers_max + 1
        self._window_top.addstr(0, x, self.SIZE_CAPTION)

    def finish(self):
        self._window_top.clear()

    def refresh(self):
        self._window_top.refresh()

    def clear(self):
        self._window_top.clear()


class BottomBar(object):
    SEARCH_CAPTION = 'Search: '
    SEARCH_IN_PROGRESS_CAPTION = 'Searching...'
    LOADING_MORE_RESULTS_CAPTION = 'Loading more results...'
    FETCHING_TORRENT_CAPTION = 'Fetching torrent...'

    def __init__(self, screen, search_fn, start_search):
        self._screen = screen
        self._search_fn = search_fn

        self._search_term = ''
        self._search_input = False
        self._search_cur_pos = 0
        self._delta_cur_pos = 0

        self._search_history = self._load_search_history()
        self._add_to_search_history(start_search)

        self._search_history_index = len(self._search_history)
        self._search_history_temp = ''

        h, _ = self._screen.getmaxyx()
        self._window_bottom = self._screen.subwin(
            1, 0, self._screen.getmaxyx()[0] - 1, 1
        )
        self._window_bottom.bkgd(' ', curses.color_pair(1) | curses.A_BOLD)

    @property
    def search_input(self):
        return self._search_input

    def resize(self):
        h, _ = self._screen.getmaxyx()
        self._window_bottom = self._screen.subwin(1, 0, h - 1, 1)
        self._window_bottom.bkgd(' ', curses.color_pair(1) | curses.A_BOLD)
        self._window_bottom.addstr(
            0, 0, ' ' * (self._screen.getmaxyx()[1] - 2)
        )
        self._window_bottom.clear()
        self._window_bottom.refresh()

    def finish(self):
        self._window_bottom.clear()
        self._save_search_history(self._search_history)

    def clear(self):
        self._window_bottom.clear()

    def process_key(self, key):
        if key != -1 and chr(key) in '%s%s%s ' % (
                string.ascii_letters, string.digits, string.punctuation
        ):
            i = len(self._search_term) + self._delta_cur_pos
            self._search_term = '%s%s%s' % (
                self._search_term[:i],
                chr(key),
                self._search_term[i:]
            )
            self._search_cur_pos += 1
        elif key in [curses.KEY_ENTER, ord("\n")]:
            self._add_to_search_history(self._search_term)
            self.set_search(False)
            self._search_fn(self._search_term)
        elif key == curses.KEY_BACKSPACE:
            i = len(self._search_term) + self._delta_cur_pos
            if i > 0:
                self._search_term = '%s%s' % (
                    self._search_term[:i - 1],
                    self._search_term[i:]
                )
                self._search_cur_pos -= 1
        elif key == curses.KEY_DC:
            i = len(self._search_term) + self._delta_cur_pos
            if i < len(self._search_term):
                self._search_term = '%s%s' % (
                    self._search_term[:i],
                    self._search_term[i + 1:]
                )
                self._delta_cur_pos += 1
        elif key == curses.KEY_LEFT:
            if self._delta_cur_pos + len(self._search_term) > 0:
                self._delta_cur_pos -= 1
        elif key == curses.KEY_RIGHT:
            if self._delta_cur_pos + len(self._search_term) < len(
                    self._search_term
            ):
                self._delta_cur_pos += 1
        elif key == curses.ascii.ESC:
            self.set_search(False)
        elif key == curses.KEY_UP:
            if self._search_history_index > 0:
                if self._search_history_index == len(self._search_history):
                    self._search_history_temp = self._search_term
                self._search_history_index -= 1
                self._search_term = self._search_history[
                    self._search_history_index
                ]
                self._delta_cur_pos = 0
        elif key == curses.KEY_DOWN:
            if self._search_history_index < len(self._search_history) - 1:
                self._search_history_index += 1
                self._search_term = self._search_history[
                    self._search_history_index
                ]
            else:
                self._search_history_index = len(self._search_history)
                self._search_term = self._search_history_temp
            self._delta_cur_pos = 0

    def set_search(self, is_search=True):
        if is_search:
            self._search_term = ''
        self._search_input = is_search

    def draw(self):
        self._window_bottom.addstr(
            0, 0, ' ' * (self._screen.getmaxyx()[1] - 2)
        )
        self._search_cur_pos = len(BottomBar.SEARCH_CAPTION) + \
                               len(self._search_term) + self._delta_cur_pos

        if self.search_input:
            self._window_bottom.addstr(
                0, 0, '%s%s' % (self.SEARCH_CAPTION, self._search_term)
            )
            self._window_bottom.move(0, self._search_cur_pos)
        self._window_bottom.refresh()

    def refresh(self):
        self._window_bottom.refresh()

    def set_search_in_progress(self):
        self._window_bottom.clear()
        self._window_bottom.addstr(0, 0, self.SEARCH_IN_PROGRESS_CAPTION)
        self._window_bottom.refresh()

    def set_loading_more_progress(self):
        self._window_bottom.clear()
        self._window_bottom.addstr(0, 0, self.LOADING_MORE_RESULTS_CAPTION)
        self._window_bottom.refresh()

    def set_fetching_magnet_url(self):
        self._window_bottom.clear()
        self._window_bottom.addstr(0, 0, self.FETCHING_TORRENT_CAPTION)
        self._window_bottom.refresh()

    def _load_search_history(self):
        history = []
        if os.path.exists(cfg.CFG_HISTORY_FILE):
            with open(cfg.CFG_HISTORY_FILE) as f:
                for line in f.readlines():
                    history.append(line.rstrip('\n'))

        return deque(history, cfg.HISTORY_MAX_LENGTH)

    def _save_search_history(self, search_history):
        with open(cfg.CFG_HISTORY_FILE, 'w') as f:
            for r in search_history:
                f.write('%s\n' % r)

    def _add_to_search_history(self, search_term):
        if search_term and (
                not self._search_history or
                self._search_term != self._search_history[-1]
        ):
            self._search_history.append(search_term)

        self._search_history_index = len(self._search_history)


class EngineSelectionWindow(object):
    WINDOW_CAPTION = 'Engine Selection'
    BUTTON_OK_CAPTION = 'OK'
    BUTTON_SAVE_CAPTION = 'SAVE'

    def __init__(self, screen, downloader):
        self._screen = screen
        self._active_engines = downloader.engines
        self._all_engines = downloader.all_engines

        self.active = False
        self._window = None
        self._position = 0

    def set_active(self, active):
        self.active = active
        if active:
            self._position = 0
            self._mk_window()
        else:
            self._window.erase()
            self._window = None

    def draw(self):
        if self.active:
            win_w = self._mk_win_w()

            self._window.erase()
            self._window.box()
            self._window.addstr(
                0, (win_w - len(self.WINDOW_CAPTION)) // 2, self.WINDOW_CAPTION
            )
            self._window.refresh()

            items = self._all_engines.copy()
            items.append(self.BUTTON_OK_CAPTION)
            items.append(self.BUTTON_SAVE_CAPTION)
            for i, item in enumerate(items):
                if i == self._position:
                    mode = curses.A_REVERSE
                else:
                    mode = curses.A_NORMAL

                if type(item) is str:
                    if len(self.WINDOW_CAPTION) % 2 == 0:
                        sp = ' ' * ((win_w - len(item) - 3) // 2)
                        cap = '[%s%s%s]' % (sp, item, sp)
                    else:
                        sp = ' ' * ((win_w - len(item) - 4) // 2)
                        cap = '[%s %s%s]' % (sp, item, sp)
                    self._window.addstr(i + 1, 1, cap, mode)
                else:
                    s = self._mk_item_caption(item)
                    cap = '%s%s' % (s, ' ' * (win_w - 3 - len(s)))
                    self._window.addstr(i + 1, 1, cap, mode)
                    s = '[X]' if item in self._active_engines.keys() else '[ ]'
                    self._window.addstr(i + 1, win_w - 4, s, mode)

    def process_key(self, key):
        if key in (curses.ascii.ESC, ord('p')):
            self.set_active(False)
        elif key == curses.KEY_UP:
            self._navigate(-1)
        elif key == curses.KEY_DOWN:
            self._navigate(1)
        elif key in (curses.KEY_ENTER, ord("\n"), 32):
            if self._position == len(self._all_engines):
                self.set_active(False)
            elif self._position == len(self._all_engines) + 1:
                cfg.SEARCH_ENGINES = [e.NAME for e in self._active_engines]
                cfg.write_cfg()
                self.set_active(False)
            else:
                self._select_engine()

    def resize(self):
        if self._window:
            self._mk_window()

    def _select_engine(self):
        if len(self._active_engines) > 1 or \
                self._all_engines[self._position] != \
                tuple(self._active_engines.keys())[0]:
            e = self._all_engines[self._position]
            if e in self._active_engines.keys():
                del self._active_engines[e]
            else:
                self._active_engines[e] = e()

    def _navigate(self, n):
        self._position += n
        if self._position < 0:
            self._position = 0
        elif self._position >= len(self._all_engines) + 2:
            self._position = len(self._all_engines) + 2 - 1

    def _mk_win_w(self):
        a = [len(self._mk_item_caption(e)) + 7 for e in self._all_engines]
        a.append(len(self.WINDOW_CAPTION) + 2)
        return max(a)

    def _mk_window(self):
        h, w = self._screen.getmaxyx()
        win_h = len(self._all_engines) + 4

        win_w = self._mk_win_w()

        self._window = self._screen.subwin(
            win_h, win_w, (h - win_h) // 2, (w - win_w) // 2
        )
        self._window.bkgd(' ', curses.color_pair(1) | curses.A_BOLD)

    def _mk_item_caption(self, item):
        return ('%s (%s)' % (
            item.BASE_URL, item.NAME
        )).split('//')[1].replace('www.', '')


class ItemWindow(object):
    SORT_ORIGIN = 0
    SORT_SEEDERS = 1
    SORT_LEECHERS = 2
    SORT_SIZE = 3
    SORT_DESC = 0
    SORT_ASC = 1

    LOAD_MORE_CAPTION = 'Load More...'
    SEARCHING_FOR_CAPTION = 'Searching for "%s"...'
    START_SEARCH_CAPTION = 'Press \'/\' to search.'
    NOTHING_FOUND_CAPTION = ' Nothing found, press \'/\' to search. '

    class LoadMoreItem(SearchResult):
        def __init__(self):
            super().__init__(None, ItemWindow.LOAD_MORE_CAPTION, 0, 0, 0, '')

    def __init__(
            self,
            screen,
            load_more_results_fn,
            fetch_magnet_fn,
            start_search_fn,
            open_torrent_link_fn,
            engine_selection_fn
    ):

        self._screen = screen
        self._load_more_results_fn = load_more_results_fn
        self._fetch_magnet_fn = fetch_magnet_fn
        self._start_search_fn = start_search_fn
        self._open_torrent_link_fn = open_torrent_link_fn
        self._engine_selection_fn = engine_selection_fn

        self._window = screen.subwin(0, 0)
        self._window.keypad(1)

        self._items = None
        self._position = 0
        self._draw_start_index = 0

        self._current_sort = self.SORT_SEEDERS
        self._current_sort_type = self.SORT_DESC

    @property
    def items(self):
        return self._items

    def process_key(self, key):
        do_exit = False
        h, w = self._screen.getmaxyx()
        if key in (curses.KEY_ENTER, ord("\n")):
            if self._items:
                item = self._items[self._position]
                if type(item) is self.LoadMoreItem:
                    self._load_more_results_fn()
                else:
                    self._fetch_magnet_fn(item)
        if key == 32:
            if self._items:
                item = self._items[self._position]
                if type(item) is not self.LoadMoreItem:
                    self._open_torrent_link_fn(item)
        elif key == curses.KEY_UP:
            if self._items:
                self._navigate(-1)
        elif key == curses.KEY_DOWN:
            if self._items:
                self._navigate(1)
        elif key == curses.KEY_PPAGE:
            if self._items:
                self._navigate(-(h - 2) // 2)
        elif key == curses.KEY_NPAGE:
            if self._items:
                self._navigate((h - 2) // 2)
        elif key == curses.ascii.ESC:
            do_exit = True
        elif key == ord('/'):
            self._start_search_fn()
        elif key == ord('a'):
            self._items = self._sort_items(self._items, self.SORT_ORIGIN)
        elif key == ord('s'):
            self._items = self._sort_items(self._items, self.SORT_SEEDERS)
        elif key == ord('d'):
            self._items = self._sort_items(self._items, self.SORT_LEECHERS)
        elif key == ord('f'):
            self._items = self._sort_items(self._items, self.SORT_SIZE)
        elif key == ord('m'):
            self._load_more_results_fn()
        elif key == ord('p'):
            self._engine_selection_fn()

        return do_exit

    def set_items(self, search_results, append=False):
        if append:
            lm = self._items.pop()
            self._items.extend(
                self._sort_items(search_results, self.SORT_SEEDERS, True)
            )
            self._items.append(lm)
        else:
            self._position = 0
            self._items = self._sort_items(
                search_results, self.SORT_SEEDERS, True
            )
            if self._items:
                self._items.append(self.LoadMoreItem())

    def draw(self,
             h,
             w,
             no_len,
             title_length,
             source_max,
             seeders_max,
             leechers_max,
             size_max,
             start_search
             ):
        if self._items:
            if h - 2 < len(self._items):
                max_items = min((h - 2, len(self._items)))
                if self._position >= self._draw_start_index + max_items \
                        or self._position < len(self._items) - max_items and \
                        self._draw_start_index > self._position:
                    if self._position - max_items + 1 > 0:
                        if self._position + 1 == self._draw_start_index:
                            self._draw_start_index = max((0, self._position))
                        else:
                            self._draw_start_index = max(
                                (0, self._position - max_items + 1)
                            )
                    else:
                        self._draw_start_index = max((0, self._position))
            else:
                max_items = len(self._items)
                self._draw_start_index = 0
        else:
            max_items = 0

        if self._items is None:
            if start_search:
                m = self.SEARCHING_FOR_CAPTION % start_search
            else:
                m = self.START_SEARCH_CAPTION
            self._window.addstr(h // 2, w // 2 - len(m) // 2, m)
        elif not self._items:
            m = self.NOTHING_FOUND_CAPTION
            self._window.addstr(h // 2, w // 2 - len(m) // 2, m)
        else:
            for index in range(max_items):
                item = self._items[index + self._draw_start_index]
                if index + self._draw_start_index == self._position:
                    mode = curses.A_REVERSE
                else:
                    mode = curses.A_NORMAL

                if type(item) is self.LoadMoreItem:
                    msg = '%s' % item.name
                    color = curses.color_pair(2)
                else:
                    color = curses.color_pair(3)
                    msg = "%s. %s|%s|%s|%s|%s" % (
                        str(index + self._draw_start_index + 1).zfill(
                            no_len
                        ),
                        self._format_field(item.name, title_length, True),
                        self._format_field(item.origin.NAME, source_max),
                        self._format_field(str(item.seeders), seeders_max),
                        self._format_field(str(item.leechers), leechers_max),
                        self._format_field(item.size, size_max)
                    )
                msg = msg + ' ' * (w - len(msg) - 1)
                self._window.addstr(1 + index, 1, msg, mode | color)

            if max_items < h - 2:
                for i in range(1, (h - 2) - max_items):
                    self._window.addstr(max_items + i, 1, ' ' * (w - 1))

    def resize(self):
        self._window = self._screen.subwin(0, 0)
        self._window.keypad(1)
        self._window.clear()

        if self._items:
            h, _ = self._screen.getmaxyx()
            max_items = min((h - 2, len(self._items)))
            if self._position > len(self._items) - max_items:
                self._draw_start_index = self._position - max_items

    def refresh(self):
        self._window.refresh()

    def finish(self):
        self._window.clear()

    def _navigate(self, n):
        self._position += n
        if abs(n) == 1:
            if self._position < 0:
                self._position = len(self._items) - 1
            elif self._position >= len(self._items):
                self._position = 0
        else:
            if self._position < 0:
                self._position = 0
            elif self._position >= len(self._items):
                self._position = len(self._items) - 1

    def _sort_items(self, items, sort_type, ignore_reverse=False):
        if items and len(items) > 0:
            if type(items[-1]) is self.LoadMoreItem:
                lmi = items.pop()
            else:
                lmi = None

            if sort_type == self.SORT_SEEDERS:
                items = sorted(
                    items,
                    key=lambda it: it.seeders,
                    reverse=self._sort_reverse(self.SORT_SEEDERS)
                            or ignore_reverse
                )
            elif sort_type == self.SORT_LEECHERS:
                items = sorted(
                    items,
                    key=lambda it: it.leechers,
                    reverse=self._sort_reverse(self.SORT_LEECHERS)
                            or ignore_reverse
                )
            elif sort_type == self.SORT_SIZE:
                items = sorted(
                    items,
                    key=lambda it: it.size_b,
                    reverse=self._sort_reverse(self.SORT_SIZE)
                            or ignore_reverse
                )
            elif sort_type == self.SORT_ORIGIN:
                items = sorted(
                    items,
                    key=lambda it: it.origin.NAME,
                    reverse=not self._sort_reverse(self.SORT_ORIGIN)
                            or ignore_reverse
                )

            if lmi:
                items.append(lmi)

            if ignore_reverse:
                self._current_sort_type = self.SORT_DESC

        return items

    def _sort_reverse(self, sort_type):
        if self._current_sort == sort_type:
            if self._current_sort_type == self.SORT_DESC:
                self._current_sort_type = self.SORT_ASC
                return False
            else:
                self._current_sort_type = self.SORT_DESC
                return True
        else:
            self._current_sort_type = self.SORT_DESC
            self._current_sort = sort_type
            return True

    def _format_field(self, s, max_, dots=False):
        if not s:
            s = 'None'

        ln = len(s)
        if dots:
            if ln > max_ - 3:
                s = '%s...' % s[:max_ - 3]
            elif ln < max_ + 3:
                s = s + ' ' * (max_ - ln)
        else:
            if ln < max_:
                s = (' ' * (max_ - ln)) + s

        return s

    def clear(self):
        self._window.clear()


class App(object):
    def __init__(self, screen, search='', loop=None):
        self._screen = screen
        self._start_search_str = search
        self._loop = asyncio.get_event_loop() or loop

        self._downloader = DlFacade()

        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_GREEN)
        curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_WHITE, curses.COLOR_BLACK)

        self._top_bar = TopBar(self._screen)
        self._bottom_bar = BottomBar(
            self._screen, self._do_search, self._start_search_str
        )
        self._item_window = ItemWindow(
            screen,
            self._load_more_results,
            self._fetch_magnet_url,
            self._bottom_bar.set_search,
            self._open_torrent_link,
            self._create_selection_win
        )
        self._engine_selection_win = EngineSelectionWindow(
            self._screen, self._downloader
        )

        self._display()

    def _display(self):
        while True:
            self._draw()

            if self._start_search_str:
                self._item_window.set_items(
                    self._loop.run_until_complete(
                        self._downloader.fetch_pages(self._start_search_str)
                    )
                )
                self._item_window.clear()
                self._start_search_str = None
                key = -1
            else:
                key = self._screen.getch()

            if self._bottom_bar.search_input:
                self._bottom_bar.process_key(key)
            elif self._engine_selection_win.active:
                self._engine_selection_win.process_key(key)
                if not self._engine_selection_win.active:
                    self._item_window.clear()
            else:
                do_exit = self._item_window.process_key(key)
                if do_exit:
                    break

            if key == curses.KEY_RESIZE:
                self._process_screen_resize()

        self._top_bar.finish()
        self._bottom_bar.finish()
        self._item_window.finish()

        curses.doupdate()

    def _draw(self):
        h, w = self._screen.getmaxyx()

        a = [len(str(i.seeders)) for i in self._item_window.items] \
            if self._item_window.items else []
        a.append(len(TopBar.SEED_CAPTION))
        seeders_max = max(a)

        a = [len(str(i.leechers)) for i in self._item_window.items] \
            if self._item_window.items else []
        a.append(len(TopBar.LEECH_CAPTION))
        leechers_max = max(a)

        a = [len(i.size) for i in self._item_window.items] \
            if self._item_window.items else []
        a.append(len(TopBar.SIZE_CAPTION))
        size_max = max(a)

        a = [
            len(i.origin.NAME) for i in self._item_window.items
            if type(i) is not ItemWindow.LoadMoreItem
        ] if self._item_window.items else []
        a.append(len(TopBar.SOURCE_CAPTION))
        source_max = max(a)

        no_len = len(str(len(self._item_window.items))) \
            if self._item_window.items else 0

        title_length = w - (
                seeders_max + leechers_max + size_max + no_len + source_max +
                len(TopBar.TITLE_CAPTION) + 2
        )
        if title_length < 0:
            title_length = 0

        self._item_window.draw(
            h,
            w,
            no_len,
            title_length,
            source_max,
            seeders_max,
            leechers_max,
            size_max,
            self._start_search_str
        )

        if self._item_window.items:
            self._top_bar.draw(
                w, no_len, source_max, seeders_max, leechers_max, size_max
            )
        else:
            self._top_bar.resize()

        self._bottom_bar.draw()

        self._engine_selection_win.draw()

        self._top_bar.refresh()
        self._item_window.refresh()
        self._bottom_bar.refresh()

        curses.doupdate()

    def _do_search(self, search_term):
        self._item_window.clear()
        self._top_bar.clear()
        self._bottom_bar.set_search_in_progress()
        self._item_window.refresh()

        self._item_window.set_items(
            self._loop.run_until_complete(
                self._downloader.fetch_pages(search_term)
            )
        )

    def _process_screen_resize(self):
        self._top_bar.resize()
        self._bottom_bar.resize()
        self._item_window.resize()
        self._engine_selection_win.resize()

    def _create_selection_win(self):
        self._engine_selection_win.set_active(True)

    def _load_more_results(self):
        self._bottom_bar.set_loading_more_progress()
        items = self._loop.run_until_complete(
            self._downloader.fetch_pages(None)
        )
        self._item_window.set_items(items, True)

    def _fetch_magnet_url(self, item):
        self._bottom_bar.set_fetching_magnet_url()

        if item.magnet_url:
            core.run_torrent_client(item.magnet_url)
        else:
            ml = asyncio.run(self._downloader.get_magnet_url(item))
            if ml:
                core.run_torrent_client(ml)
            else:
                # TODO magnurl error
                pass

    def _open_torrent_link(self, item):
        core.open_torrent_link('%s%s' % (item.origin.BASE_URL, item.link))



import json

from bs4 import BeautifulSoup

from tordl.core import BaseDl, SearchResult


class SolidTorrents(BaseDl):
    NAME = 'Solid'
    BASE_URL = 'https://solidtorrents.net'
    SEARCH_URL = \
        '%s/api/v1/search?sort=seeders&q=%s&category=all&skip=%s&fuv=yes' % \
        (BASE_URL, '%s', '%s')

    def _mk_search_url(self, expression):
        return self.SEARCH_URL % (
            expression, str((self._current_index - 1) * 20)
        )

    def _process_search(self, response):
        result = []
        try:
            results = json.loads(response)['results']
            for o in results:
                swarm = o['swarm']
                result.append(
                    SearchResult(
                        type(self),
                        o['title'],
                        '/search?q=%s' %
                        self._current_search,
                        swarm['seeders'],
                        swarm['leechers'],
                        self._hr_size(o['size']),
                        o['magnet']
                    )
                )
        except Exception:
            pass
        return result

    def _mk_magnet_url(self, link):
        pass

    def _process_magnet_link(self, response):
        pass

    def _hr_size(self, size):
        if size < 1024:
            return '%dB' % size
        else:
            size = size / 1024
            if size < 1024:
                return '%.2fKB' % size
            else:
                size = size / 1024
                if size < 1024:
                    return '%.2fMB' % size
                else:
                    size = size / 1024
                    if size < 1024:
                        return '%.2fGB' % size
                    else:
                        size = size / 1024
                        return '%.2fTB' % size


class KickAssTorrents(BaseDl):
    NAME = 'KAT'
    BASE_URL = 'https://kickasss.to'
    # BASE_URL = 'https://katcr.to'
    SEARCH_URL = '%s/search/%s/%s/' % (BASE_URL, '%s', '%s')

    def _mk_search_url(self, expression):
        return self.SEARCH_URL % (expression, str(self._current_index))

    def _process_search(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        result = []
        try:
            trs = bs.findAll('tr', class_='odd')
            trs.extend(bs.findAll('tr', class_='even'))
            for tr in trs:
                link = tr.find(class_='cellMainLink').attrs['href']
                size = tr.find(class_='nobr center').text.replace('\n', '')
                seeders = int(tr.find(class_='green center').text)
                leechers = int(tr.find(class_='red lasttd center').text)
                name = tr.find(class_='cellMainLink').text.replace('\n', ''). \
                    strip()
                result.append(
                    SearchResult(
                        type(self), name, link, seeders, leechers, size
                    )
                )
        except Exception:
            pass
        return result

    def _mk_magnet_url(self, link):
        return '%s%s' % (self.BASE_URL, link)

    def _process_magnet_link(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        try:
            m = bs.find('a', class_='kaGiantButton').attrs['href']
            return m
        except Exception:
            return None


class GloTorrents(BaseDl):
    NAME = 'Glo'
    BASE_URL = 'https://www.gtdb.to'
    SEARCH_URL = '%s/search_results.php?search=%s&sort=seeders&' \
                 'order=desc&page=%s' % (BASE_URL, '%s', '%s')

    def _mk_search_url(self, expression):
        return self.SEARCH_URL % (expression, str(self._current_index - 1))

    def _process_search(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        result = []
        try:
            trs = bs.find(class_='ttable_headinner').findAll('tr')[1:]
            for tr in trs:
                c2 = tr.findAll(class_='ttable_col2')
                if len(c2) > 1:
                    c1 = tr.findAll(class_='ttable_col1')
                    a = c2[0].findAll('a')[1]
                    name = a.attrs['title']
                    link = a.attrs['href']
                    magnet_url = c2[1].find('a').attrs['href']
                    size = c1[2].text
                    seeders = int(c2[2].find('b').text.replace(',', ''))
                    leechers = int(c1[3].find('b').text.replace(',', ''))
                    result.append(
                        SearchResult(
                            type(self),
                            name,
                            link,
                            seeders,
                            leechers,
                            size,
                            magnet_url
                        )
                    )
        except Exception:
            pass
        return result

    def _mk_magnet_url(self, link):
        pass

    def _process_magnet_link(self, response):
        pass


class ZooqleCom(BaseDl):
    NAME = 'Zooqle'
    BASE_URL = 'https://zooqle.com'
    SEARCH_URL = '%s/search?pg=%s&q=%s&s=ns&v=t&sd=a' % (BASE_URL, '%s', '%s')

    def __init__(self):
        super().__init__()

        self._s_cls = [
            'progress-bar smaller prog-green prog-l',
            'progress-bar smaller prog-yellow prog-l',
            'progress-bar smaller prog-red prog-l',
        ]
        self._l_cls = [
            'progress-bar smaller prog-green prog-r',
            'progress-bar smaller prog-yellow prog-r',
            'progress-bar smaller prog-red prog-r',
        ]

    def _mk_search_url(self, expression):
        return self.SEARCH_URL % (str(self._current_index), expression)

    def _process_search(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        result = []
        try:
            trs = bs.find(
                class_='table table-condensed table-torrents vmiddle'). \
                      findAll('tr')[1:]
            for tr in trs:
                name = tr.find(class_='text-trunc text-nowrap')
                a = name.find('a')
                name = a.text
                link = a.attrs['href']
                magnet_url = tr.find(align='left').find('ul'). \
                    findAll('li')[1].find('a').attrs['href']
                try:
                    size = tr.find(class_='progress-bar prog-blue prog-l').text
                except Exception:
                    size = '0B'
                seeders = self._get_from_cls(tr, self._s_cls)
                leechers = self._get_from_cls(tr, self._l_cls)
                result.append(
                    SearchResult(
                        type(self),
                        name,
                        link,
                        seeders,
                        leechers,
                        size,
                        magnet_url
                    )
                )

        except Exception:
            pass

        return result

    def _get_from_cls(self, tr, cls):
        for c in cls:
            try:
                seeders = int(tr.find(class_=c).text)
                return seeders
            except Exception:
                pass

        return 0

    def _mk_magnet_url(self, link):
        pass

    def _process_magnet_link(self, response):
        pass


class TpbParty(BaseDl):
    NAME = 'TPB'
    BASE_URL = 'https://tpb.party'
    SEARCH_URL = '%s/search/%s/%s/99/0' % (BASE_URL, '%s', '%s')

    def _mk_search_url(self, expression):
        return self.SEARCH_URL % (expression, str(self._current_index))

    def _process_search(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        result = []
        try:
            trs = bs.find('table', id='searchResult').findAll('tr')[1:]
            for tr in trs:
                a = tr.find(class_='detName').find('a')
                name = a.text
                link = '/%s' % '/'.join(a.attrs['href'].split('/')[3:])
                tds = tr.findAll('td')
                magnet_url = tds[1].findAll('a')[1].attrs['href']
                seeders = int(tds[2].text)
                leeches = int(tds[3].text)
                size = tr.find(class_='detDesc').text.split(',')[1]
                size = size.replace('Size ', '').replace('i', '')
                result.append(
                    SearchResult(
                        type(self),
                        name,
                        link,
                        seeders,
                        leeches,
                        size,
                        magnet_url
                    )
                )
        except Exception:
            pass

        return result

    def _mk_magnet_url(self, link):
        pass

    def _process_magnet_link(self, response):
        pass


class LimeTorrents(BaseDl):
    NAME = 'Lime'
    BASE_URL = 'https://www.limetorrents.info'
    SEARCH_URL = '%s/search/all/%s/seeds/%s/' % (BASE_URL, '%s', '%s')

    def _mk_search_url(self, expression):
        return self.SEARCH_URL % (expression, str(self._current_index))

    def _mk_magnet_url(self, link):
        return '%s%s' % (self.BASE_URL, link)

    def _process_search(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        result = []
        try:
            t = bs.find(class_='table2')
            trs = t.findAll('tr')[1:]
            for tr in trs:
                a = tr.find(class_='tt-name').findAll('a')[1]
                name = a.string
                link = a.attrs['href']
                size = tr.findAll(class_='tdnormal')[1].string
                seeders = int(tr.find(class_='tdseed').string.replace(',', ''))
                leechers = int(
                    tr.find(class_='tdleech').string.replace(',', '')
                )
                result.append(
                    SearchResult(
                        type(self), name, link, seeders, leechers, size
                    )
                )
        except Exception:
            pass

        return result

    def _process_magnet_link(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        try:
            return bs.findAll(class_='csprite_dltorrent')[2].attrs['href']
        except Exception:
            pass

        return None


class Dl1337xto(BaseDl):
    NAME = '1337x'
    BASE_URL = 'https://1337x.to'
    SEARCH_URL = '%s/search/%s/%s/' % (BASE_URL, '%s', '%s')

    def _mk_search_url(self, expression):
        return self.SEARCH_URL % (expression, str(self._current_index))

    def _mk_magnet_url(self, magnet_page_link):
        return '%s%s' % (self.BASE_URL, magnet_page_link)

    def _process_search(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        result = []
        try:
            trs = bs.find(
                class_='table-list table table-responsive table-striped'
            ).find('tbody').findAll('tr')
            for tr in trs:
                tds = tr.findAll('td')

                a = tds[0].findAll('a')[1]
                name = a.string
                link = a.attrs['href']

                seeders = int(tds[1].string)
                leechers = int(tds[2].string)
                size = '%sB' % tds[4].text.split('B')[0]

                result.append(
                    SearchResult(
                        type(self), name, link, seeders, leechers, size
                    )
                )
        except Exception:
            pass

        return result

    def _process_magnet_link(self, response):
        bs = BeautifulSoup(response, features='html.parser')
        try:
            x = bs.find(class_='box-info torrent-detail-page vpn-info-wrap')
            if not x:
                x = bs.find(
                    class_=
                    'box-info torrent-detail-page series-torrent vpn-info-wrap'
                )
            x = x.findAll('div')[1].find('div').find('ul').find('li').find('a')
            return x.attrs['href']
        except Exception:
            return None
